# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Data for tests and documentation with BIFROST."""

_version = "3"


def _make_pooch():
    import pooch

    return pooch.create(
        path=pooch.os_cache("ess/bifrost"),
        env="ESS_DATA_DIR",
        base_url="https://public.esss.dk/groups/scipp/ess/bifrost/{version}/",
        version=_version,
        registry={
            "BIFROST_20240914T053723.h5": "md5:0f2fa5c9a851f8e3a4fa61defaa3752e",
            "computed_energy_data_simulated.h5": "blake2b:3c398443cb85c8294d283c7212255bc695c2520f2332c2c99d041a0760b6bcbb9937e19bcd8a498daf306d279c88d2ea911c510c1ce3b3a7f1e6b7e54022a943",  # noqa: E501
            "computed_energy_data_simulated_5x2.h5": "blake2b:d9d5e785a08e14d9c3416cf04db89a8c6f2fae3c0bae27bf0e73e8e5d492b4ca406e6578a935fa9f72dd9199dc15536409f614791fee6899c4265fe5d31e2706",  # noqa: E501
            "BIFROST-simulation-tof-lookup-table.h5": "blake2b:682021920a355f789da37b18029719fe20569d86db26cdaf5f3d916d2f76f9360907960ba86903be4cab489d39f1b6f9f265f3a4ab3f82c5e095afa4a2c456af",  # noqa: E501
        },
        retry_if_failed=3,
    )


_pooch = _make_pooch()


def get_path(name: str) -> str:
    """
    Return the path to a data file bundled with ess.bifrost.

    This function only works with example data and cannot handle
    paths to custom files.
    """
    return _pooch.fetch(name)


def simulated_elastic_incoherent_with_phonon() -> str:
    """Simulated data for elastic incoherent scattering including a phonon."""
    return get_path("BIFROST_20240914T053723.h5")


def tof_lookup_table_simulation() -> str:
    """Time-of-flight lookup table for the simulated BIFROST data.

    This table was computed with `tof <https://github.com/scipp/tof>`_
    using `Create a time-of-flight lookup table for BIFROST
    <../../user-guide/bifrost/bifrost-make-tof-lookup-table.rst>`_
    with ``NumberOfSimulatedNeutrons = 5_000_000``.
    """
    return get_path("BIFROST-simulation-tof-lookup-table.h5")


def computed_energy_data_simulated() -> str:
    """Energy and momentum transfer computed from the simulated BIFROST data."""
    return get_path("computed_energy_data_simulated.h5")


def computed_energy_data_simulated_5x2() -> str:
    """Energy and momentum transfer computed from the simulated BIFROST data.

    This reference was computed with 10 detectors forming a 5x2 grid
    (arc=5, channel=2).
    """
    return get_path("computed_energy_data_simulated_5x2.h5")


__all__ = [
    "computed_energy_data_simulated",
    "computed_energy_data_simulated_5x2",
    "get_path",
    "simulated_elastic_incoherent_with_phonon",
    "tof_lookup_table_simulation",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Utilities for computing real neutron time-of-flight."""

from ess.reduce import time_of_flight as reduce_time_of_flight
from ess.reduce.time_of_flight.types import DetectorLtotal
from ess.spectroscopy.types import (
    DataGroupedByRotation,
    DetectorData,
    DetectorTofData,
    PulseStrideOffset,
    RunType,
    TimeOfFlightLookupTable,
)


def detector_time_of_flight_data(
    sample_data: DataGroupedByRotation[RunType],
    lookup: TimeOfFlightLookupTable,
    ltotal: DetectorLtotal[RunType],
    pulse_stride_offset: PulseStrideOffset,
) -> DetectorTofData[RunType]:
    """
    Convert the time-of-arrival data to time-of-flight data using a lookup table.

    The output data will have a time-of-flight coordinate.

    This is a wrapper around
    :func:`ess.reduce.time_of_flight.detector_time_of_flight_data`
    for different input types.
    """
    return reduce_time_of_flight.eto_to_tof.detector_time_of_flight_data(
        detector_data=DetectorData[RunType](sample_data),
        lookup=lookup,
        ltotal=ltotal,
        pulse_stride_offset=pulse_stride_offset,
    )


providers = (detector_time_of_flight_data,)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Handling of final parameters / secondary spectrometer."""

import numpy as np
import scipp as sc
from scippneutron.conversion.beamline import (
    BeamAlignedUnitVectors,
)

from ess.spectroscopy.types import (
    Analyzer,
    DataAtSample,
    DataGroupedByRotation,
    PulsePeriod,
    RunType,
    SecondarySpecCoordTransformGraph,
)


def sample_analyzer_vector(
    sample_position: sc.Variable,
    analyzer_position: sc.Variable,
    analyzer_transform: sc.Variable,
    detector_position: sc.Variable,
) -> sc.Variable:
    """Determine the sample to analyzer-reflection-point vector per detector element

    Note
    ----
    The shapes of the analyzer position and orientation should be self-consistent
    and will likely be 1:1. There is expected to be multiple detector element positions
    per analyzer which can be represented as an additional dimension compared to
    the analyzer shapes.

    Parameters
    ----------
    sample_position:
        The (probably scalar) sample position, typically vector (0, 0, 0).
    analyzer_position:
        The nominal center of the central analyzer blade *surface*.
    analyzer_transform:
        The transform that places and rotates the analyzer in lab space.
        Only the rotation is extracted to identify the crystal y-axis.
    detector_position:
        The position of the detector element.

    Returns
    -------
    :
        The vector from the sample position to the interaction point on the analyzer
        for each detector element.
    """
    # Scipp does not distinguish between coordinates and directions, so we need to do
    # some extra legwork to ensure we can apply the orientation transformation
    # _and_ obtain a dimensionless direction vector
    o = sc.vector([0, 0, 0], unit=analyzer_transform.unit)
    y = sc.vector(
        [0, 1, 0], unit=analyzer_transform.unit
    )  # and y perpendicular to the scattering plane
    yhat = analyzer_transform * y - analyzer_transform * o
    yhat /= sc.norm(yhat)

    sample_analyzer_center_vector = analyzer_position - sample_position

    sample_detector_vector = detector_position - sample_position
    sd_out_of_plane = sc.dot(sample_detector_vector, yhat)
    # the sample-detector vector is the sum of sample-analyzer,
    # analyzer-detector-center, the out-of-plane vector
    analyzer_detector_center_vector = (
        sample_detector_vector - sample_analyzer_center_vector - sd_out_of_plane * yhat
    )

    # TODO Consider requiring that dot(analyzer_position-sample_position, yhat) is zero?

    sample_analyzer_center_distance = sc.norm(sample_analyzer_center_vector)

    analyzer_detector_center_distance = sc.norm(analyzer_detector_center_vector)

    # similar-triangles give the out-of-plane analyzer reflection point distance
    sa_out_of_plane = (
        sample_analyzer_center_distance
        / (sample_analyzer_center_distance + analyzer_detector_center_distance)
        * sd_out_of_plane
    )

    return sample_analyzer_center_vector + sa_out_of_plane * yhat


def detector_geometric_a4(
    *,
    sample_analyzer_vector: sc.Variable,
    beam_aligned_unit_vectors: BeamAlignedUnitVectors,
) -> sc.Variable:
    """Calculate the scattering angle from the incident beam to each detector element

    Parameters
    ----------
    sample_analyzer_vector:
        The per detector element analyzer interaction position, as determined
        by :func:`sample_analyzer_vector`.
    beam_aligned_unit_vectors:
        Unit vectors defining the coordinate system aligned with the incident beam
        and gravity.
        See :func:`scippneutron.conversion.beamline.beam_aligned_unit_vectors`.

    Returns
    -------
    :
        The per detector element scattering angle, a4, in degrees.
    """
    lab_x = beam_aligned_unit_vectors['beam_aligned_unit_x']
    lab_z = beam_aligned_unit_vectors['beam_aligned_unit_z']
    x = sc.dot(lab_x, sample_analyzer_vector)
    z = sc.dot(lab_z, sample_analyzer_vector)
    return sc.atan2(y=x, x=z).to(unit='deg')


def analyzer_detector_vector(
    sample_position: sc.Variable,
    sample_analyzer_vector: sc.Variable,
    detector_position: sc.Variable,
) -> sc.Variable:
    """Calculate the analyzer-detector vector"""
    analyzer_position = sample_position + sample_analyzer_vector.to(
        unit=sample_analyzer_vector.unit
    )
    return detector_position - analyzer_position.to(unit=detector_position.unit)


def final_wavenumber(
    sample_analyzer_vector: sc.Variable,
    analyzer_detector_vector: sc.Variable,
    analyzer_dspacing: sc.Variable,
) -> sc.Variable:
    r"""Find the wave number of the neutrons reflected to each detector-element.

    The wave number and wavelength are inversely proportional with

    .. math::

        \text{wavelength} = 2 * \pi / \text{wavenumber}

    Parameters
    ----------
    sample_analyzer_vector:
        The vector from the sample to the analyzer interaction point for each
        detector element.
    analyzer_detector_vector:
        The vector from the analyzer interaction point to its detector element,
        for each detector element.
    analyzer_dspacing:
        The lattice plane spacing of the analyzer crystal.

    Returns
    -------
    :
        The magnitude of the reflected neutron wave vector for each detector element.
    """
    # law of Cosines gives the scattering angle based on distances:
    l_sa = sc.norm(sample_analyzer_vector)
    l_ad = sc.norm(analyzer_detector_vector)
    l_diff = sc.norm(sample_analyzer_vector + analyzer_detector_vector)
    # 2 theta is measured from the direction S-A, so the internal angle is
    # (pi - 2 theta) and the normal law of Cosines is modified accordingly to be
    # -cos(2 theta) instead of cos(pi - 2 theta)
    cos2theta = (l_diff**2 - l_sa**2 - l_ad**2) / (2 * l_sa * l_ad)

    # law of Cosines gives the Bragg reflected wavevector magnitude
    return (
        2 * np.pi / sc.sqrt(2 - 2 * cos2theta) / analyzer_dspacing.to(unit='angstrom')
    )


def final_energy(final_wavenumber: sc.Variable) -> sc.Variable:
    """Converts (final) wave number to (final) energy"""
    from scipp.constants import hbar, neutron_mass

    return ((hbar**2 / 2 / neutron_mass) * final_wavenumber**2).to(unit='meV')


def final_wavevector(
    sample_analyzer_vector: sc.Variable, final_wavenumber: sc.Variable
) -> sc.Variable:
    """Constructs the final wave vector from its direction and magnitude"""
    return (
        sample_analyzer_vector
        / sc.norm(sample_analyzer_vector)
        * final_wavenumber.to(unit='1/angstrom')
    )


def secondary_flight_path_length(
    sample_analyzer_vector: sc.Variable,
    analyzer_detector_vector: sc.Variable,
) -> sc.Variable:
    """Returns the path-length-distance between the sample and each detector element"""
    ad = sc.norm(analyzer_detector_vector)
    sa = sc.norm(sample_analyzer_vector).to(unit=ad.unit)
    return sa + ad


def secondary_flight_time(
    L2: sc.Variable, final_wavenumber: sc.Variable
) -> sc.Variable:
    """Calculates the most-likely time-of-flight between the sample and each pixel"""
    from scipp.constants import hbar, neutron_mass

    velocity = final_wavenumber * (hbar / neutron_mass)
    return sc.to_unit(L2 / velocity, 'ms', copy=False)


def secondary_spectrometer_coordinate_transformation_graph(
    analyzer: Analyzer[RunType],
) -> SecondarySpecCoordTransformGraph[RunType]:
    """Return a coordinate transformation graph for the secondary spectrometer.

    Parameters
    ----------
    analyzer:
        Data group with analyzer parameters.

    Returns
    -------
    :
        Coordinate transformation graph for the secondary spectrometer.
        The graph captures the relevant parameters of ``analyzer``.
    """
    from scippneutron.conversion.beamline import beam_aligned_unit_vectors

    return SecondarySpecCoordTransformGraph[RunType](
        {
            "a4": detector_geometric_a4,
            "analyzer_dspacing": lambda: analyzer["dspacing"],
            "analyzer_position": lambda: analyzer["position"],
            "analyzer_transform": lambda: analyzer["transform"],
            "beam_aligned_unit_vectors": beam_aligned_unit_vectors,
            "detector_position": "position",
            "sample_analyzer_vector": sample_analyzer_vector,
            "analyzer_detector_vector": analyzer_detector_vector,
            "final_energy": final_energy,
            "final_wavenumber": final_wavenumber,
            "final_wavevector": final_wavevector,
            "L2": secondary_flight_path_length,
            "secondary_flight_time": secondary_flight_time,
        }
    )


def move_time_to_sample(
    data: DataGroupedByRotation[RunType], pulse_period: PulsePeriod
) -> DataAtSample[RunType]:
    """Return the events with the event_time_offset coordinate offset to time at sample.

    Parameters
    ----------
    data:
        Data array with "event_time_offset" and "secondary_flight_time" coordinates.
    pulse_period:
        Duration of a neutron pulse.

    Returns
    -------
    :
        A shallow copy of ``data`` where the "event_time_offset" coordinate has been
        shifted to the time at the sample.
    """
    offset = data.coords['secondary_flight_time'].to(
        unit=data.bins.coords['event_time_offset'].unit, copy=False
    )
    time = data.bins.coords['event_time_offset'] - offset
    time %= pulse_period.to(unit=time.unit)
    return DataAtSample[RunType](
        data
        # These are the detector positions and they no longer match the time:
        .drop_coords(('position', 'x_pixel_offset', 'y_pixel_offset'))
        # Use bins.assign_coords to avoid modifying the original data:
        .bins.assign_coords(event_time_offset=time)
    )


providers = (
    secondary_spectrometer_coordinate_transformation_graph,
    move_time_to_sample,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Coordinate conversions for single crystal diffraction with BIFROST."""

from scippneutron.conversion.graph import beamline as beamline_graphs
from scippneutron.conversion.graph import tof as tof_graphs

from ess.spectroscopy.indirect.conversion import (
    rotate_to_sample_table_momentum_transfer,
)
from ess.spectroscopy.types import (
    DetectorCountsWithQ,
    DetectorTofData,
    ElasticCoordTransformGraph,
    GravityVector,
    RunType,
)


def single_crystal_coordinate_transformation_graph(
    gravity: GravityVector,
) -> ElasticCoordTransformGraph:
    """Return the coordinate transformation graph for single crystal diffraction."""
    base = tof_graphs.elastic_Q_vec(start='tof')
    base['lab_momentum_transfer'] = base['Q_vec']
    return ElasticCoordTransformGraph(
        {
            **beamline_graphs.beamline(scatter=True),
            **base,
            'gravity': lambda: gravity,
            'sample_table_momentum_transfer': rotate_to_sample_table_momentum_transfer,
        }
    )


def convert_tof_to_q(
    with_tof: DetectorTofData[RunType],
    *,
    graph: ElasticCoordTransformGraph,
) -> DetectorCountsWithQ[RunType]:
    """Convert ToF to Q."""
    transformed = with_tof.transform_coords(
        ['a3', 'sample_table_momentum_transfer'],
        graph=graph,
        keep_intermediate=False,
        keep_inputs=False,
        keep_aliases=False,
        rename_dims=False,  # because otherwise, it would rename a3 -> Q
    )
    return DetectorCountsWithQ[RunType](transformed)


providers = (single_crystal_coordinate_transformation_graph, convert_tof_to_q)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

"""Input / output for indirect spectroscopy."""

from __future__ import annotations

import h5py
import numpy
import scippnexus
from scipp import DataArray

from ..types import EnergyData, NXspeFileName, NXspeFileNames, SampleRun


def to_nxspe(events: EnergyData[SampleRun], base: NXspeFileName) -> NXspeFileNames:
    """Take events, which have been binned in incident wavelength and have monitor
    counts per bin, and output one NXspe file per setting

    Parameters
    ----------
    events: scipp.DataArray
        The events binned in (setting, event_id, incident_wavelength) with
        at least 'monitor', 'a3', and 'theta' bin coordinates.
        The events and the bins must also have an 'incident_wavelength' coordinate.
    base: str | Path
        The filename base used to produce each NXspe filename.

    Returns
    -------
    :
        The list of filenames containing the NXspe data. If there are N settings in the
        input DataArray, N filenames are returned. The names are of the form

        .. code-block::

            {base}_{i+1:0{ceil(log10(N+1))}d}.nxspe
    """
    from pathlib import Path

    dim = 'setting'
    length = len(str(events.sizes[dim] + 1))
    files = []
    base = Path(base)
    parent = base.parent
    if not parent.exists():
        parent.mkdir(parents=True)

    for i in range(events.sizes[dim]):
        ev = events[dim, i]
        fn = str(base) + '_' + f'{i + 1}'.rjust(length, '0') + '.nxspe'
        files.append(NXspeFileName(fn))
        _to_one_nxspe(ev, fn)
    return NXspeFileNames(files)


def _make_group(group: h5py.Group) -> scippnexus.Group:
    return scippnexus.Group(group, definitions=scippnexus.base_definitions())


def _lambda_to_ei(incident_wavelength):
    from scipp.constants import Planck, neutron_mass

    return ((Planck / incident_wavelength) ** 2 / neutron_mass / 2).to(unit='meV')


def _to_one_nxspe(events: DataArray, filename: str):
    """Use scippnexus to create the NXspe file"""
    import scipp as sc
    from scippnexus import (
        NXcollection,
        NXdata,
        NXentry,
        NXfermi_chopper,
        NXinstrument,
        NXsample,
    )

    from .conservation import energy_transfer

    observations = events.copy()
    ef = events.coords['final_energy']
    observations *= sc.sqrt(events.bins.coords['incident_energy'] / ef)

    # Adding null events requires replicating all coordinates of real events.
    # We don't necessarily use all present event coordinates, so remove any we won't use
    targets = [
        'energy_transfer',
        'incident_energy',
        'incident_wavelength',
        'final_energy',
    ]
    for coord in [x for x in events.bins.coords if x not in targets]:
        del observations.bins.coords[coord]
    # And provide a means to calculate the rest from bin (not event) information
    # we are sure incident_wavelength is a bin coordinate already, so can skip it.
    graph = {
        'incident_energy': _lambda_to_ei,
        'energy_transfer': energy_transfer,
    }
    # # Adding zero-weight observations, i.e., null events, works now, but involves
    # # a memory-copy of the data array and its bin structure. Since it isn't strictly
    # # necessary for exporting NXspe, it can be skipped for now.
    # observations = add_null_observations(observations, targets, graph)

    # combine the per bin intensities and normalize by monitor counts
    # Note, applying this normalization to the _events_ would require splitting
    # the bin monitor counts between the bin events.
    # we need to ignore the monitor uncertainty for the time being
    normalize_by = sc.values(events.coords['monitor'])

    observations = observations.hist().transform_coords(targets, graph=graph)

    if observations.variances is None:
        observations.variances = observations.values  # correct for counting statistics
        observations.variances[observations.values == 0] = 1
    # the transform_coords above renamed the 'incident_wavelength' dimension
    # to 'energy_transfer' ... so we need to do the same to normalize_by or else
    # scipp tries to broadcast when it doesn't need to.
    observations.data = observations.data / normalize_by.rename_dims(
        incident_wavelength='energy_transfer'
    )

    psi = observations.coords['a3']
    polar = observations.coords['theta']
    azimuthal = sc.zeros(sizes=polar.sizes, unit='deg', dtype=polar.dtype)
    azimuthal_width = sc.full(sizes=polar.sizes, unit='deg', value=2.0)
    polar_width = sc.full(sizes=polar.sizes, unit='deg', value=0.1)
    distance = sc.full(sizes=polar.sizes, unit='m', value=3.0, dtype=polar.dtype)
    data = observations.data
    if observations.data.variances is not None:
        error = numpy.sqrt(observations.data.variances)
    else:
        error = 0 * observations.data.values
    energy_transfer = observations.coords['energy_transfer']
    incident_energy = observations.coords['incident_energy']
    final_energy = observations.coords['final_energy']

    with h5py.File(filename, mode='w') as f:
        # make / in the file
        root = _make_group(f)
        # it _must_ contain an NXentry group, called [anything which is allowed]?
        # with two fields and five subgroups required
        entry = root.create_class('entry', NXentry)
        # the name of  the author program
        entry.create_field('program_name', sc.scalar('essspectroscopy'))
        # and the NXDL schema information -- currently version 3.1
        definition = entry.create_field('definition', sc.scalar('NXSPE'))
        definition.attrs['version'] = '3.1'

        # the entry group also contains five subgroups

        # the NXcollection group must contain three fields
        nxinfo = entry.create_class('NXSPE_info', NXcollection)
        nxinfo.create_field('fixed_energy', final_energy)
        nxinfo.create_field('ki_over_kf_scaling', sc.scalar(True))
        nxinfo.create_field('psi', psi)

        # the NXdata group has 8 required fields
        nxdata = entry.create_class('data', NXdata)
        nxdata.create_field('azimuthal', azimuthal)
        nxdata.create_field('azimuthal_width', azimuthal_width)
        nxdata.create_field('polar', polar)
        nxdata.create_field('polar_width', polar_width)
        nxdata.create_field('distance', distance)
        nxdata.create_field('data', data)
        nxdata.create_field('error', error)
        nxdata.create_field('energy', energy_transfer)
        # Actually more useful extensions to NXspe for an instrument like BIFROST
        nxdata.create_field('final_energy', final_energy)
        nxdata.create_field('incident_energy', incident_energy)

        # the NXinstrument group has one required field and one required group
        instrument = entry.create_class('instrument', NXinstrument)
        instrument.create_field('name', sc.scalar('SIMBIFROST'))
        fermi = instrument.create_class('fermi_chopper', NXfermi_chopper)
        fermi.create_field('energy', sc.scalar(numpy.nan, unit='meV'))

        # and the NXsample group has three required fields
        sample = entry.create_class('sample', NXsample)
        sample.create_field('rotation_angle', psi)
        sample.create_field('seblock', sc.scalar(""))
        sample.create_field('temperature', sc.scalar(numpy.nan, unit='K'))


providers = (to_nxspe,)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Detector handling for BIFROST."""

import scipp as sc
import scippnexus as snx

from ess.spectroscopy.types import (
    BeamlineWithSpectrometerCoords,
    CalibratedDetector,
    DetectorPositionOffset,
    NeXusComponent,
    NeXusTransformation,
    RunType,
)

from .types import ArcNumber


def arc_number(
    beamline: BeamlineWithSpectrometerCoords[RunType],
) -> ArcNumber[RunType]:
    """Calculate BIFROST arc index number from pixel final energy

    The BIFROST analyzers are each set to diffract an
    energy in the set (2.7, 3.2, 3.8, 4.4. 5.0) meV.
    This energy is only valid for the central point of the center
    tube of the associated detector triplet. All other pixels
    will have a final energy slightly higher or lower.

    This function assigns the closest arc number indexing the
    ordered set above.

    Parameters
    ----------
    beamline:
        A data array with a 'final_energy' coordinate which is the
        per-pixel (or event) final neutron energy.

    Returns
    -------
    :
        The arc index of the analyzer from which the neutron scattered
    """
    minimum = sc.scalar(2.7, unit='meV')
    step = sc.scalar(0.575, unit='meV')
    final_energy = beamline.coords['final_energy']
    return ArcNumber[RunType](sc.round((final_energy - minimum) / step).to(dtype='int'))


def get_calibrated_detector_bifrost(
    detector: NeXusComponent[snx.NXdetector, RunType],
    *,
    transform: NeXusTransformation[snx.NXdetector, RunType],
    offset: DetectorPositionOffset[RunType],
) -> CalibratedDetector[RunType]:
    """Extract the data array corresponding to a detector's signal field.

    The returned data array includes coords and masks pertaining directly to the
    signal values array, but not additional information about the detector.
    The data array is reshaped to the logical detector shape.

    This function is specific to BIFROST and differs from the generic
    :func:`ess.reduce.nexus.workflow.get_calibrated_detector` in that it does not
    fold the detectors into logical dimensions because the files already contain
    the detectors in the correct shape.

    Parameters
    ----------
    detector:
        Loaded NeXus detector.
    transform:
        Transformation that determines the detector position.
    offset:
        Offset to add to the detector position.

    Returns
    -------
    :
        Detector data.
    """

    from ess.reduce.nexus.types import DetectorBankSizes
    from ess.reduce.nexus.workflow import get_calibrated_detector

    da = get_calibrated_detector(
        detector=detector,
        transform=transform,
        offset=offset,
        # The detectors are folded in the file, no need to do that here.
        bank_sizes=DetectorBankSizes({}),
    )
    da = da.rename(dim_0='tube', dim_1='length')

    # Calculate arc and channel for this triplet based on detector_number layout.
    # BIFROST detector_number ordering is (arc, tube, channel, pixel).
    # Each triplet contains 3 tubes of 100 pixels for a single arc-channel pair.
    det_num = da.coords['detector_number']['tube', 0]['length', 0].value
    pixels_per_tube = 100
    tubes_per_channel = 3
    channels_per_arc = 9
    pixels_per_arc = pixels_per_tube * tubes_per_channel * channels_per_arc  # 2700

    # detector_number is 1-indexed
    idx = det_num - 1
    arc = idx // pixels_per_arc
    remainder = idx % pixels_per_arc
    channel = (remainder % (pixels_per_tube * channels_per_arc)) // pixels_per_tube

    da.coords['arc'] = sc.index(arc)
    da.coords['channel'] = sc.index(channel)

    return CalibratedDetector[RunType](da)


def merge_triplets(
    *triplets: sc.DataArray,
) -> sc.DataArray:
    """Merge BIFROST detector triplets into a single data array.

    This function folds the triplets into (arc, channel) dimensions based on
    the scalar 'arc' and 'channel' coordinates assigned to each triplet. If the
    triplets form a regular rectangular subset of the full 5x9 detector array,
    they will be folded into those dimensions. Otherwise, they are concatenated
    along a 'triplet' dimension.

    Parameters
    ----------
    triplets:
        Data arrays to merge. Each must have scalar 'arc' and 'channel' coordinates.

    Returns
    -------
    :
        Input data arrays either folded into (arc, channel) dimensions or
        stacked along the "triplet" dimension.
    """
    if len(triplets) == 0:
        raise ValueError("At least one triplet is required")

    # Extract arc and channel from scalar coordinates
    arc_channel_pairs = [
        (triplet.coords['arc'].value, triplet.coords['channel'].value)
        for triplet in triplets
    ]

    # Sort triplets by (arc, channel)
    sorted_indices = sorted(range(len(triplets)), key=lambda i: arc_channel_pairs[i])
    sorted_triplets = [triplets[i] for i in sorted_indices]
    sorted_pairs = [arc_channel_pairs[i] for i in sorted_indices]

    # Check if the pairs form a regular rectangular grid
    unique_arcs = sorted({pair[0] for pair in sorted_pairs})
    unique_channels = sorted({pair[1] for pair in sorted_pairs})

    # Check if we have a complete rectangular subset
    expected_pairs = [
        (arc, channel) for arc in unique_arcs for channel in unique_channels
    ]
    concatenated = sc.concat(sorted_triplets, dim='triplet')
    if sorted_pairs == expected_pairs:
        # We have a regular grid, fold it
        return concatenated.fold(
            dim='triplet',
            sizes={'arc': len(unique_arcs), 'channel': len(unique_channels)},
        )

    # Fall back to simple concatenation if not a regular grid
    return concatenated


providers = (arc_number, get_calibrated_detector_bifrost)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Handling of initial parameters / primary spectrometer."""

from __future__ import annotations

from ess.spectroscopy.types import (
    PrimarySpecCoordTransformGraph,
    RunType,
)


def primary_spectrometer_coordinate_transformation_graph() -> (
    PrimarySpecCoordTransformGraph[RunType]
):
    """Return a coordinate transformation graph for the primary spectrometer.

    Returns
    -------
    :
        Coordinate transformation graph for the primary spectrometer.
    """
    # For the incident beam, the original implementation here used the guides to
    # determine a more accurate estimate of the path length.
    # See function `primary_path_length` in
    # commit 929ef7f97e00a1e26c254fd5f08c8a3346255970
    # The result differs by <1mm from the straight line distance.
    # This should be well below the measurement accuracy.
    # So we use the simpler straight line distance here.

    from scippneutron.conversion.beamline import L1, straight_incident_beam

    return PrimarySpecCoordTransformGraph[RunType](
        {
            "incident_beam": straight_incident_beam,
            "L1": L1,
        }
    )


providers = (primary_spectrometer_coordinate_transformation_graph,)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
# ruff: noqa: E402, F401

"""BIFROST data reduction."""

import importlib.metadata

from .detector import providers
from .io import nexus
from .workflow import BifrostSimulationWorkflow, BifrostWorkflow

try:
    __version__ = importlib.metadata.version("essspectroscopy")
except importlib.metadata.PackageNotFoundError:
    __version__ = "0.0.0"

del importlib

__all__ = ['BifrostSimulationWorkflow', 'BifrostWorkflow', 'nexus', 'providers']



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Domain types for spectroscopy."""

from typing import Any, NewType, TypeVar

import sciline
import scipp as sc

from ess.reduce import time_of_flight
from ess.reduce.nexus import types as reduce_t

# NeXus types

CalibratedBeamline = reduce_t.CalibratedBeamline
CalibratedDetector = reduce_t.CalibratedDetector
DetectorData = reduce_t.DetectorData
DetectorPositionOffset = reduce_t.DetectorPositionOffset
GravityVector = reduce_t.GravityVector
Filename = reduce_t.Filename
MonitorData = reduce_t.MonitorData
NeXusClass = reduce_t.NeXusClass
NeXusComponentLocationSpec = reduce_t.NeXusComponentLocationSpec
NeXusComponent = reduce_t.NeXusComponent
NeXusData = reduce_t.NeXusData
NeXusDetectorName = reduce_t.NeXusDetectorName
NeXusFileSpec = reduce_t.NeXusFileSpec
NeXusMonitorName = reduce_t.NeXusName
NeXusTransformation = reduce_t.NeXusTransformation
Position = reduce_t.Position
PreopenNeXusFile = reduce_t.PreopenNeXusFile


SampleRun = reduce_t.SampleRun
VanadiumRun = reduce_t.VanadiumRun

FrameMonitor0 = reduce_t.FrameMonitor0
FrameMonitor1 = reduce_t.FrameMonitor1
FrameMonitor2 = reduce_t.FrameMonitor2
FrameMonitor3 = reduce_t.FrameMonitor3

# Type vars

RunType = TypeVar("RunType", SampleRun, VanadiumRun)
MonitorType = TypeVar(
    "MonitorType",
    FrameMonitor0,
    FrameMonitor1,
    FrameMonitor2,
    FrameMonitor3,
)

# Time-of-flight types

DetectorTofData = time_of_flight.DetectorTofData
MonitorTofData = time_of_flight.MonitorTofData
MonitorLtotal = time_of_flight.MonitorLtotal
PulseStride = time_of_flight.PulseStride
PulseStrideOffset = time_of_flight.PulseStrideOffset
PulsePeriod = time_of_flight.PulsePeriod
TimeOfFlightLookupTable = time_of_flight.TimeOfFlightLookupTable
TimeOfFlightLookupTableFilename = time_of_flight.TimeOfFlightLookupTableFilename

# Custom types


class Analyzer(sciline.Scope[RunType, sc.DataGroup[Any]], sc.DataGroup[Any]): ...


class Analyzers(sciline.Scope[RunType, sc.DataGroup[Any]], sc.DataGroup[Any]): ...


class BeamlineWithSpectrometerCoords(
    sciline.Scope[RunType, sc.DataArray], sc.DataArray
): ...


class DataAtSample(sciline.Scope[RunType, sc.DataArray], sc.DataArray): ...


class DataGroupedByRotation(sciline.Scope[RunType, sc.DataArray], sc.DataArray): ...


class DetectorCountsWithQ(sciline.Scope[RunType, sc.DataArray], sc.DataArray): ...


ElasticCoordTransformGraph = NewType('ElasticCoordTransformGraph', dict)


class EnergyData(sciline.Scope[RunType, sc.DataArray], sc.DataArray): ...


InelasticCoordTransformGraph = NewType('InelasticCoordTransformGraph', dict)


class SampleAngle(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Rotation angle of the sample, possibly as a function of time.

    For BIFROST, this is angle "a3".
    """


class InstrumentAngle(sciline.Scope[RunType, sc.DataArray], sc.DataArray):
    """Rotation angle of the instrument, possibly as a function of time.

    For BIFROST, this is angle "a4".
    """


MonitorCoordTransformGraph = NewType('MonitorCoordTransformGraph', dict)


NXspeFileName = NewType('NXspeFileName', str)
NXspeFileNames = NewType('NXspeFileNames', list[NXspeFileName])


class PrimarySpecCoordTransformGraph(sciline.Scope[RunType, dict], dict): ...


class SecondarySpecCoordTransformGraph(sciline.Scope[RunType, dict], dict): ...


class WavelengthMonitor(
    sciline.Scope[RunType, MonitorType, sc.DataArray], sc.DataArray
): ...



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""Utilities for computing real neutron time-of-flight for indirect geometry."""

from collections.abc import Iterable

import sciline

from ess.reduce import time_of_flight as reduce_time_of_flight
from ess.reduce.time_of_flight.types import DetectorLtotal

from ..types import (
    DataAtSample,
    DetectorData,
    DetectorTofData,
    MonitorCoordTransformGraph,
    MonitorData,
    MonitorLtotal,
    MonitorTofData,
    MonitorType,
    PulseStrideOffset,
    RunType,
    TimeOfFlightLookupTable,
)


def TofWorkflow(
    *,
    run_types: Iterable[sciline.typing.Key],
    monitor_types: Iterable[sciline.typing.Key],
) -> sciline.Pipeline:
    workflow = reduce_time_of_flight.GenericTofWorkflow(
        run_types=run_types,
        monitor_types=monitor_types,
    )
    for provider in providers:
        workflow.insert(provider)
    return workflow


def detector_time_of_flight_data(
    sample_data: DataAtSample[RunType],
    lookup: TimeOfFlightLookupTable,
    pulse_stride_offset: PulseStrideOffset,
) -> DetectorTofData[RunType]:
    """
    Convert the time-of-arrival data to time-of-flight data using a lookup table.

    The output data will have a time-of-flight coordinate.

    This is a wrapper around
    :func:`ess.reduce.time_of_flight.detector_time_of_flight_data`
    for indirect geometry spectrometers.
    """
    result = reduce_time_of_flight.eto_to_tof.detector_time_of_flight_data(
        detector_data=DetectorData[RunType](sample_data),
        lookup=lookup,
        ltotal=DetectorLtotal(sample_data.coords['L1']),
        pulse_stride_offset=pulse_stride_offset,
    )
    # This is time-of-flight at the sample.
    result.bins.coords['sample_tof'] = result.bins.coords.pop('tof')
    del result.bins.coords['event_time_zero']
    return result


def monitor_time_of_flight_data(
    monitor_data: MonitorData[RunType, MonitorType],
    lookup: TimeOfFlightLookupTable,
    ltotal: MonitorLtotal[RunType, MonitorType],
    pulse_stride_offset: PulseStrideOffset,
) -> MonitorTofData[RunType, MonitorType]:
    """
    Convert the time-of-arrival data to time-of-flight data using a lookup table.

    The output data will have a time-of-flight coordinate.

    This is a wrapper around
    :func:`ess.reduce.time_of_flight.monitor_time_of_flight_data`
    for indirect geometry spectrometers.
    """
    result = reduce_time_of_flight.eto_to_tof.monitor_time_of_flight_data(
        monitor_data=monitor_data.rename(t='tof'),
        lookup=lookup,
        ltotal=ltotal,
        pulse_stride_offset=pulse_stride_offset,
    )
    return result


def compute_monitor_ltotal(
    monitor_data: MonitorData[RunType, MonitorType],
    coord_transform_graph: MonitorCoordTransformGraph,
) -> MonitorLtotal[RunType, MonitorType]:
    """Compute the path length from the source to the monitor."""
    return MonitorLtotal[RunType, MonitorType](
        monitor_data.transform_coords(
            'Ltotal',
            graph=coord_transform_graph,
            keep_intermediate=False,
            keep_aliases=False,
            rename_dims=False,
        ).coords['Ltotal']
    )


providers = (
    compute_monitor_ltotal,
    detector_time_of_flight_data,
    monitor_time_of_flight_data,
)
"""Providers for time-of-flight calculation on indirect geometry spectrometers.

The providers here override the default providers of
:class:`ess.reduce.time_of_flight.GenericTofWorkflow`
to customize the workflow for indirect geometry spectrometers.
"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""NeXus input/output for BIFROST."""

import scipp as sc
import scippnexus as snx

from ess.reduce.nexus import load_all_components, open_component_group
from ess.reduce.nexus.types import NeXusAllLocationSpec, NeXusLocationSpec
from ess.spectroscopy.types import (
    Analyzer,
    Analyzers,
    InstrumentAngle,
    NeXusClass,
    NeXusComponentLocationSpec,
    NeXusFileSpec,
    RunType,
    SampleAngle,
)


# See https://github.com/scipp/essreduce/issues/98
def moderator_class_for_source() -> NeXusClass[snx.NXsource]:
    """Select NXmoderator as the source."""
    return NeXusClass[snx.NXsource](snx.NXmoderator)


def load_sample_angle(
    file_spec: NeXusFileSpec[RunType],
) -> SampleAngle[RunType]:
    return SampleAngle[RunType](_load_experiment_parameter(file_spec, "a3"))


def load_instrument_angle(
    file_spec: NeXusFileSpec[RunType],
) -> InstrumentAngle[RunType]:
    return InstrumentAngle[RunType](_load_experiment_parameter(file_spec, "a4"))


def _load_experiment_parameter(
    file_spec: NeXusFileSpec[RunType], param_name: str
) -> sc.DataArray:
    with open_component_group(
        NeXusLocationSpec(filename=file_spec.value),
        nx_class=snx.NXparameters,
        parent_class=snx.NXentry,
    ) as group:
        return group[param_name][()]['value']


def load_analyzers(file_spec: NeXusFileSpec[RunType]) -> Analyzers[RunType]:
    """Load all analyzers in a NeXus file."""
    return Analyzers[RunType](
        load_all_components(
            NeXusAllLocationSpec(filename=file_spec.value),
            nx_class=snx.NXcrystal,
        )
    )


def _get_analyzer_for_detector_name(
    detector_name: str, analyzers: Analyzers[RunType]
) -> Analyzers[RunType]:
    detector_index = int(detector_name.split('_', 1)[0])
    analyzer_index = str(detector_index - 2)
    for name, analyzer in analyzers.items():
        if name.startswith(analyzer_index):
            return analyzer
    raise RuntimeError(f"No analyzer found for detector {detector_name}")


def analyzer_for_detector(
    analyzers: Analyzers[RunType],
    detector_location: NeXusComponentLocationSpec[snx.NXdetector, RunType],
) -> Analyzer[RunType]:
    """Extract the analyzer for a given detector.

    Note
    ----
    Depends heavily on the names of components being preceded by an instrument index,
    and the analyzer and detector components being separated in index by 2.
    If either condition changes, this function will need to be modified.

    Parameters
    ----------
    analyzers:
        Data group of loaded analyzers.
    detector_location:
        The location of an NXdetector in the NeXus file.
        The analyzer is identified based on this location.

    Returns
    -------
    :
        The analyzer for the given detector triplet.
        Only a subset of fields is returned.
    """
    if detector_location.component_name is None:
        raise ValueError("Detector component name is None")
    analyzer = snx.compute_positions(
        _get_analyzer_for_detector_name(detector_location.component_name, analyzers),
        store_transform='transform',
    )
    return Analyzer[RunType](
        sc.DataGroup(
            dspacing=analyzer['d_spacing'],
            position=analyzer['position'],
            transform=analyzer['transform'],
        )
    )


providers = (
    analyzer_for_detector,
    load_analyzers,
    load_instrument_angle,
    load_sample_angle,
    moderator_class_for_source,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc
import scipp.testing
import scippnexus as snx

from ess import bifrost
from ess.bifrost.data import (
    simulated_elastic_incoherent_with_phonon,
    tof_lookup_table_simulation,
)
from ess.bifrost.live import CutAxis, CutAxis1, CutAxis2, arc_energy, cut
from ess.spectroscopy.types import (
    EnergyData,
    Filename,
    NeXusDetectorName,
    SampleRun,
    TimeOfFlightLookupTable,
)


@pytest.fixture(scope='module')
def simulation_detector_names() -> list[NeXusDetectorName]:
    with snx.File(simulated_elastic_incoherent_with_phonon()) as f:
        names = list(f['entry']['instrument'][snx.NXdetector].keys())
    return names[:10]  # First 10 detectors form a 5x2 grid (arc=5, channel=2)


@pytest.fixture
def energy_data(
    simulation_detector_names: list[NeXusDetectorName],
) -> EnergyData[SampleRun]:
    workflow = bifrost.BifrostSimulationWorkflow(simulation_detector_names)
    workflow[Filename[SampleRun]] = simulated_elastic_incoherent_with_phonon()
    workflow[TimeOfFlightLookupTable] = sc.io.load_hdf5(tof_lookup_table_simulation())
    return workflow.compute(EnergyData[SampleRun])


def test_cut_along_q_norm_and_energy_transfer_preserves_counts(
    energy_data: EnergyData[SampleRun],
) -> None:
    # Define cut axes for |Q| and energy transfer
    axis_1 = CutAxis(
        output='|Q|',
        fn=lambda sample_table_momentum_transfer: sc.norm(
            sample_table_momentum_transfer
        ),
        bins=sc.linspace(dim='|Q|', start=0.0, stop=3.0, num=50, unit='1/angstrom'),
    )
    axis_2 = CutAxis(
        output='E',
        fn=lambda energy_transfer: energy_transfer,
        bins=sc.linspace(dim='E', start=-10.0, stop=10.0, num=50, unit='meV'),
    )

    cut_data = cut(
        energy_data,
        axis_1=CutAxis1(axis_1),
        axis_2=CutAxis2(axis_2),
        arc_energy=arc_energy(),
    )

    # Verify no counts were lost during the cut
    total_counts_before = sc.sum(energy_data.bins.size()).value
    total_counts_after = sc.sum(cut_data).value
    assert total_counts_before == total_counts_after


def test_cut_along_qx_direction_preserves_counts(
    energy_data: EnergyData[SampleRun],
) -> None:
    # Test cutting along a specific Q direction (Qx)
    axis_1 = CutAxis.from_q_vector(
        output='Qx',
        vec=sc.vector([1, 0, 0]),
        bins=sc.linspace(dim='Qx', start=-2.0, stop=2.0, num=40, unit='1/angstrom'),
    )
    axis_2 = CutAxis(
        output='E',
        fn=lambda energy_transfer: energy_transfer,
        bins=sc.linspace(dim='E', start=-5.0, stop=5.0, num=30, unit='meV'),
    )

    cut_data = cut(
        energy_data,
        axis_1=CutAxis1(axis_1),
        axis_2=CutAxis2(axis_2),
        arc_energy=arc_energy(),
    )

    # Verify no counts were lost during the cut
    total_counts_before = sc.sum(energy_data.bins.size()).value
    total_counts_after = sc.sum(cut_data).value
    assert total_counts_before == total_counts_after



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

"""McStas input for BIFROST."""

import scipp as sc
import scippnexus as snx

from ess.spectroscopy.types import (
    BeamlineWithSpectrometerCoords,
    DetectorData,
    NeXusData,
    PulsePeriod,
    RunType,
)

from ..types import McStasDetectorData


def assemble_detector_data(
    detector: BeamlineWithSpectrometerCoords[RunType],
    event_data: NeXusData[snx.NXdetector, RunType],
) -> McStasDetectorData[RunType]:
    """Custom assemble function that returns McStasDetectorData."""
    from ess.reduce.nexus.workflow import assemble_detector_data as reduce_assemble

    return McStasDetectorData[RunType](reduce_assemble(detector, event_data))


def convert_simulated_time_to_event_time_offset(
    mcstas_data: McStasDetectorData[RunType],
    pulse_period: PulsePeriod,
) -> DetectorData[RunType]:
    """Helper to make McStas simulated event data look more like real data

    McStas has the ability to track the time-of-flight from source to detector for
    every probabilistic neutron ray. This is very helpful, but unfortunately real
    instrument at ESS are not able to record the same information due to how the
    timing and data collection systems work.

    Real neutron events will record their event_time_zero most-recent-pulse reference
    time, and their event_time_offset detection time relative to that reference time.
    These two values added together give a real wall time; and information about the
    primary spectrometer is necessary to find any time-of-flight

    This function takes event data with per-event coordinate event_time_offset
    (actually McStas time-of-arrival) and converts the coordinate to be
    the time-of-arrival modulo the source repetition period.

    Notes
    -----
    If the input data has realistic event_time_offset values, this function should
    be a noop.

    Returns
    -------
    :
        A copy of the data with realistic per-event coordinate event_time_offset.
    """

    def wrap_event_time_offset(event_time_offset: sc.Variable) -> sc.Variable:
        return event_time_offset % pulse_period.to(unit=event_time_offset.unit)

    res = mcstas_data.transform_coords(
        frame_time=wrap_event_time_offset,
        rename_dims=False,
        keep_intermediate=False,
        keep_inputs=False,
    )
    return DetectorData[RunType](
        res.transform_coords(event_time_offset='frame_time', keep_inputs=False)
    )


providers = (
    assemble_detector_data,
    convert_simulated_time_to_event_time_offset,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import pytest
import sciline
import scipp as sc
import scipp.testing
import scippnexus as snx

from ess.bifrost.data import (
    simulated_elastic_incoherent_with_phonon,
    tof_lookup_table_simulation,
)
from ess.bifrost.live import BifrostQCutWorkflow, CutAxis, CutAxis1, CutAxis2, CutData
from ess.spectroscopy.types import (
    Filename,
    NeXusDetectorName,
    SampleRun,
    TimeOfFlightLookupTable,
)


@pytest.fixture(scope='module')
def simulation_detector_names() -> list[NeXusDetectorName]:
    with snx.File(simulated_elastic_incoherent_with_phonon()) as f:
        names = list(f['entry']['instrument'][snx.NXdetector].keys())
    return names[:10]  # First 10 detectors form a 5x2 grid (arc=5, channel=2)


class TestBifrostQCutWorkflow:
    @pytest.fixture
    def qcut_workflow(
        self, simulation_detector_names: list[NeXusDetectorName]
    ) -> sciline.Pipeline:
        workflow = BifrostQCutWorkflow(simulation_detector_names)
        workflow[Filename[SampleRun]] = simulated_elastic_incoherent_with_phonon()
        workflow[TimeOfFlightLookupTable] = sc.io.load_hdf5(
            tof_lookup_table_simulation()
        )
        return workflow

    def test_cut_along_q_norm_and_energy_transfer(
        self, qcut_workflow: sciline.Pipeline
    ) -> None:
        # Define cut axes for |Q| and energy transfer
        axis_1 = CutAxis(
            output='|Q|',
            fn=lambda sample_table_momentum_transfer: sc.norm(
                sample_table_momentum_transfer
            ),
            bins=sc.linspace(dim='|Q|', start=0.0, stop=3.0, num=50, unit='1/angstrom'),
        )
        axis_2 = CutAxis(
            output='E',
            fn=lambda energy_transfer: energy_transfer,
            bins=sc.linspace(dim='E', start=-10.0, stop=10.0, num=50, unit='meV'),
        )

        qcut_workflow[CutAxis1] = axis_1
        qcut_workflow[CutAxis2] = axis_2

        cut_data = qcut_workflow.compute(CutData[SampleRun])

        # Verify the structure of the result (now 3-D with arc dimension)
        assert cut_data.bins is None  # Should be histogrammed
        assert set(cut_data.dims) == {'arc', '|Q|', 'E'}
        assert cut_data.sizes['arc'] == 5
        assert cut_data.sizes['|Q|'] == 49  # num bins - 1
        assert cut_data.sizes['E'] == 49
        assert 'arc' in cut_data.coords
        assert '|Q|' in cut_data.coords
        assert 'E' in cut_data.coords

        # Check that coordinates have expected units and values
        cut_data.coords['arc'].to(unit='meV')
        cut_data.coords['|Q|'].to(unit='1/angstrom')
        cut_data.coords['E'].to(unit='meV')

        # Verify arc coordinate values
        expected_arc_energies = sc.array(
            dims=['arc'], values=[2.7, 3.2, 3.8, 4.4, 5.0], unit='meV'
        )
        sc.testing.assert_allclose(cut_data.coords['arc'], expected_arc_energies)

    def test_cut_along_qx_direction_and_energy_transfer(
        self, qcut_workflow: sciline.Pipeline
    ) -> None:
        # Test cutting along a specific Q direction (Qx)
        axis_1 = CutAxis.from_q_vector(
            output='Qx',
            vec=sc.vector([1, 0, 0]),
            bins=sc.linspace(dim='Qx', start=-2.0, stop=2.0, num=40, unit='1/angstrom'),
        )
        axis_2 = CutAxis(
            output='E',
            fn=lambda energy_transfer: energy_transfer,
            bins=sc.linspace(dim='E', start=-5.0, stop=5.0, num=30, unit='meV'),
        )

        qcut_workflow[CutAxis1] = axis_1
        qcut_workflow[CutAxis2] = axis_2

        cut_data = qcut_workflow.compute(CutData[SampleRun])

        # Verify the result structure (now 3-D with arc dimension)
        assert cut_data.bins is None
        assert set(cut_data.dims) == {'arc', 'Qx', 'E'}
        assert cut_data.sizes['arc'] == 5
        assert cut_data.sizes['Qx'] == 39
        assert cut_data.sizes['E'] == 29
        assert 'arc' in cut_data.coords
        assert 'Qx' in cut_data.coords
        assert 'E' in cut_data.coords
